import sys
import logging

entry_points = ["mastprov", "imasprov"]

if not any(entry_point in sys.argv[0] for entry_point in entry_points): # To suppress irrelevant warning if using CLI entry points
    logging.basicConfig(level=logging.INFO, format="%(message)s")

try:
    from .imasprov import ImasProv
except ModuleNotFoundError as error:
    logging.info(
        f"{error}. An IMAS installation is required to write provenance for IMAS formatted data."
    )
try:
    from .mastprov import write_provenance
except ModuleNotFoundError as error:
    logging.info(f"{error}. A UDA installation is required write provenance for MAST signals.")
except ImportError as error:
    logging.info(f"{error}. A UDA installation is required write provenance for MAST signals.")

__version__ = "1.2.0"



import argparse
import json
import os

import imas
from prov.dot import prov_to_dot
from prov.model import ProvDocument

from fusionprov import utilities


class ImasProv:
    """
    This is the class for finding the provenance information for a given
    IDS, which can then be used to produce a provenance document that adheres
    to the W3C-PROV standard.

    Attributes:
        data_ids (imas.ids): The IDS containing the data that this class will
        produce a provenance document for, e.g. 'equilibrium', 'magnetics',
        'thomson_scattering'.

        dataset_desc (imas.ids.dataset_description): The dataset_description
        IDS that corresponds to the data_ids. It is the user's responsibility to
        ensure that the two are from the same pulse file.

        dataset_fair (imas.ids.dataset_fair): The dataset_fair IDS that
        corresponds to the data_ids. It is the user's responsibility to ensure
        that the two are from the same pulse file.
    """

    def __init__(self, data_ids, dataset_desc=None, dataset_fair=None):
        self.data_ids = data_ids
        self.dataset_desc = dataset_desc
        self.dataset_fair = dataset_fair
        self.prov_doc = ProvDocument()

    def imas_uri(self):
        """
        Returns the IMAS data dictionary documentation url based on the version
        used to put the IDS. If the version is not found in the IDS, it will be
        taken from the dataset_descritpion or dataset_fair IDS. Failing that, it
        will return the url for the data_dictionary version 3.31.0
        """
        default_version = "3.31.0"
        if self.data_ids.ids_properties.version_put.data_dictionary:
            version = self.data_ids.ids_properties.version_put.data_dictionary
        elif self.dataset_desc:
            try:
                version = self.dataset_desc.ids_properties.version_put.data_dictionary
            except:
                pass
            else:
                print("Data dictionary version found in dataset_description IDS")
        elif self.dataset_fair:
            try:
                version = self.dataset_fair.ids_properties.version_put.data_dictionary
            except:
                pass
            else:
                print("Data dictionary version found in dataset_fair IDS")
        else:
            version = default_version
        uri = f"https://sharepoint.iter.org/departments/POP/CM/IMDesign/Data%20Model/CI/imas-{version}/html_documentation.html"
        return uri

    def extract_source(self):
        """
        Converts the contents of ids_properties.source field to dict if it 
        contains json. Otherwise, fills the dict using the _no_source() method.
        """
        try:
            self.data_source = json.loads(self.data_ids.ids_properties.source)
            self.data_source["source"] = self.data_ids.ids_properties.source
        except ValueError:
            print(
                "ids_properties.source not filled according to "
                "recommended template. Provenance information will be "
                "incomplete."
            )
            self.data_source = self._no_source()

        self.prov_doc.add_namespace("IMAS", self.imas_uri())
        self.prov_doc.add_namespace(
            self.data_source["database"], self.data_source["database_uri"]
        )

    def _no_source(self):
        """
        Fills data_source dict with "UNKNOWN" if the ids_properties.source field
        is not filled according to the template.
        """
        empty_source = {
            "database": "UNKNOWN",
            "database_uri": "UNKNOWN",
            "shot": "UNKNOWN",
            "run": "UNKNOWN",
            "data_uri": "UNKNOWN",
            "input_signals": "UNKNOWN",
            "source": f"SOURCE: {self.data_ids.ids_properties.source}"
        }
        return empty_source

    def create_code_agent(self):
        if self.data_ids.code.name and self.data_ids.code.repository:
            self.prov_doc.add_namespace(
                f"{self.data_ids.code.name}", f"{self.data_ids.code.repository}"
            )
        else:
            return None

        code_agent = self.prov_doc.agent(
            f"{self.data_source['database']}:{self.data_ids.code.name}",
            {
                "prov:type": "prov:SoftwareAgent",
                f"{self.data_ids.code.name}:commit": self.data_ids.code.commit,
                f"{self.data_ids.code.name}:version": self.data_ids.code.version,
                f"{self.data_ids.code.name}:repository": self.data_ids.code.repository,
                f"{self.data_ids.code.name}:parameters": self.data_ids.code.parameters,
            },
        )
        return code_agent

    def serialize_provenance(self, graph=False):
        """
        Serialize provenance document to both xml and json formats. Optionally 
        also graphically as png.
        """
        directories = ("prov-xml", "prov-json", "prov-graph")
        for directory in directories:
            if not os.path.exists(directory):
                os.makedirs(directory)

        # Generate document file in both .json and .xml
        json_file = utilities.slugify(f"{self.data_ids.__name__}_prov") + ".json"
        xml_file = utilities.slugify(f"{self.data_ids.__name__}_prov") + ".xml"
        json_path = os.path.join("prov-json", json_file)
        xml_path = os.path.join("prov-xml", xml_file)
        self.prov_doc.serialize(json_path)
        self.prov_doc.serialize(xml_path, format="xml")

        if graph:
            # Generate graphical representation
            dot = prov_to_dot(self.prov_doc)
            graph_file = utilities.slugify(f"{self.data_ids.__name__}_prov") + ".png"
            graph_path = os.path.join("prov-graph", graph_file)
            dot.write_png(graph_path)

    def prov_from_data_ids(self, graph=False):
        """
        This method will provice basic provenance for an IDS that may be
        supplied on its own without a dataset_description or dataset_fair IDS.
        """
        print(f"Generating a provenance document for {self.data_ids.__name__} ids.")

        self.extract_source()

        ids_entity = self.prov_doc.entity(f"IMAS: {self.data_ids.__name__.title()} IDS")
        data_entity = self.prov_doc.entity(
            f"{self.data_source['database']}:"
            f"shot {self.data_source['shot']}, "
            f"run {self.data_source['run']}, "
            f"{self.data_ids.__name__} data, "
            f"{self.data_source['source']}"
        )
        self.prov_doc.wasDerivedFrom(ids_entity, data_entity)

        ids_provider_agent = self.prov_doc.agent(
            f"{self.data_source['database']}:{self.data_ids.ids_properties.provider}"
        )
        code_agent = self.create_code_agent()
        self.prov_doc.wasAttributedTo(ids_entity, ids_provider_agent)

        if code_agent:
            self.prov_doc.wasAttributedTo(data_entity, code_agent)

        try:
            data_put_activity = self.prov_doc.activity(
                "IMAS:put", startTime=self.data_ids.ids_properties.creation_date,
            )
        except ValueError:
            data_put_activity = self.prov_doc.activity("IMAS:put")

        data_dict_entity = self.prov_doc.entity(
            f"IMAS:data dictionary version {self.data_ids.ids_properties.version_put.data_dictionary}"
        )
        access_layer_entity = self.prov_doc.entity(
            f"IMAS:access layer {self.data_ids.ids_properties.version_put.access_layer}"
        )
        access_layer_lang_entity = self.prov_doc.entity(
            f"IMAS:access layer language {self.data_ids.ids_properties.version_put.access_layer_language}"
        )
        self.prov_doc.used(data_put_activity, data_dict_entity)
        self.prov_doc.used(data_put_activity, access_layer_entity)
        self.prov_doc.used(data_put_activity, access_layer_lang_entity)
        self.prov_doc.wasGeneratedBy(ids_entity, data_put_activity)
        self.prov_doc.wasAssociatedWith(data_put_activity, ids_provider_agent)

        self.serialize_provenance(graph=graph)


def main():
    parser = argparse.ArgumentParser(
        description="Retrieve and document the provenance of an Interface Data "
        "Structure (IDS)."
    )
    parser.add_argument("database", help="IMAS database")
    parser.add_argument("shot", help="Shot number", type=int)
    parser.add_argument("run", help="Run number", type=int)
    parser.add_argument("ids", help="IDS name")
    parser.add_argument(
        "--graph",
        "-g",
        help="In addition to text formats, write prov in graph form as .png. "
        "Default: false",
        action="store_true",
    )
    parser.add_argument(
        "--remote",
        "-r",
        help="Request data from remote database. By default the data will be read from the local IMAS database.",
        action="store_true"
    )
    args = parser.parse_args()

    if args.remote:
        pulse = imas.ids(args.shot, args.run)
        pulse.open_public(args.database)
    else:
        user = os.environ["USER"]
        imas_version = os.environ["IMAS_VERSION"]
        pulse = imas.ids(args.shot, args.run)
        pulse.open_env(user, args.database, imas_version)

    ids = getattr(pulse, args.ids)
    ids.get()
    print("Getting IDS...")
    ids_prov = ImasProv(ids)
    ids_prov.prov_from_data_ids(graph=args.graph)



import argparse
import logging
import os
import re
import sys
from abc import ABC, abstractmethod
from collections.abc import Mapping
from itertools import zip_longest
from typing import Any, Dict

import pyuda

try:
    from mast.mast_client import ListType
except ModuleNotFoundError as error:
    logging.info(
        f"{error}. If not on the Freia cluster, provenance reports can be produced for uda signals only."
    )
    FREIA = False
else:
    FREIA = True
from prov.dot import prov_to_dot
from prov.model import ProvDocument

from fusionprov import utilities


class MastProvFactory:
    subclasses = {}

    @classmethod
    def register_subclass(cls, data_type):
        def decorator(subclass):
            cls.subclasses[data_type] = subclass
            return subclass

        return decorator

    @classmethod
    def create(cls, data_type, params):

        if data_type not in cls.subclasses:
            raise ValueError(f"Input {data_type} is not a valid data type.")

        return cls.subclasses[data_type](params)


class MastProv(ABC):
    def __init__(self, params):
        self.prov_doc = ProvDocument()
        self.client = pyuda.Client()
        self.client.set_property("get_meta", True)
        self.data = params["data"]
        self.shot = params["shot"]
        self.run = params["run"]
        self.graph = params["graph"]
        self.log_file = None

    machine = "MAST"
    namespace_uri = "https://users.mastu.ukaea.uk"
    data_location = "$MAST_DATA"

    def get_input_files(self, log_file=None):
        """
        Finds the raw or analysed datafiles that were used to produce the
        analysed data.
        """
        if not log_file:
            log_file = self.log_file
        matches = re.findall(
            r"(?<=/)[r|x][a-z][a-z][0-9]*?\.nc|(?<=/)[r|x][a-z][a-z][0-9]*?\.[0-9][0-9]|(?<=/)a[a-z][a-z][0-9]*?\.[0-9][0-9]|(?<=/)a[a-z][a-z][0-9]*?\.nc",
            log_file,
        )
        datafiles = set(matches)
        input_files = []
        for datafile in datafiles:
            if datafile[0] in ["a", "e"]:
                log = self.get_log(file_code=datafile[:3])
                input_files.append({datafile: self.get_input_files(log_file=log)})
            else:
                input_files.append(datafile)
        return input_files

    def data_ro(self):
        return "MAST INBOX"

    def get_session_leader(self):
        """
        TODO: Provide link to session information so authorised users can find
        who the session leader was.
        """
        return "SESSION LEADER"

    def get_scheduler_trigger(self):
        """
        TODO: Find a way to show why a pass was performed for run numbers
        greater than 0.
        """
        if self.run == 0:
            return "Automatic post-shot analysis run"
        else:
            return "TBD"

    def get_analysis_code_author(self):
        """
        Provides the resonsible officer for the analysis code or contact
        details for MAST Ops team.
        """
        return "MAST INBOX"

    @abstractmethod
    def build_prov(self):
        pass

    @abstractmethod
    def validate():
        pass

    def write_signal_prov(
        self,
        signal,
        signal_data_entity,
        scheduler_run_activity,
        record_shot_activity,
        scheduler_pass_trigger_agent,
        signals_done=None,
    ):
        if signals_done is None:
            signals_done = []

        if signal not in signals_done:
            if type(signal) is str:
                input_data_entity = self.prov_doc.entity(f"{MastProv.machine}:{signal}")
                self.prov_doc.wasDerivedFrom(signal_data_entity, input_data_entity)
                signal_ro_agent = self.prov_doc.agent(
                    f"{MastProv.machine}:{self.data_ro()}"
                )
                self.prov_doc.used(scheduler_run_activity, input_data_entity)
                if signal[0] in ["x", "r"]:
                    self.prov_doc.wasGeneratedBy(
                        input_data_entity, record_shot_activity
                    )
                self.prov_doc.wasAttributedTo(input_data_entity, signal_ro_agent)

            elif type(signal) is dict:
                for k, v in signal.items():
                    file_code = k[:3]
                    scheduler_run = self.get_scheduler_run(
                        file_code=file_code, log_file=self.get_log(file_code=file_code)
                    )
                    scheduler_run_activity = self.prov_doc.activity(
                        f"{MastProv.machine}:{scheduler_run[0]}",
                        startTime=scheduler_run[1],
                        endTime=scheduler_run[2],
                    )
                    self.prov_doc.wasAssociatedWith(
                        scheduler_run_activity, scheduler_pass_trigger_agent
                    )
                    analysis_code = self.get_analysis_code(file_code=file_code)
                    analysis_code_entity = self.prov_doc.entity(
                        f"{MastProv.machine}:{analysis_code}"
                    )
                    analysis_code_author = self.get_analysis_code_author()
                    analysis_code_author_agent = self.prov_doc.agent(
                        f"{MastProv.machine}:{analysis_code_author}"
                    )
                    self.prov_doc.used(scheduler_run_activity, analysis_code_entity)
                    self.prov_doc.wasAttributedTo(
                        analysis_code_entity, analysis_code_author_agent
                    )
                    self.write_signal_prov(
                        k,
                        signal_data_entity,
                        scheduler_run_activity,
                        record_shot_activity,
                        scheduler_pass_trigger_agent,
                        signals_done=signals_done,
                    )
                    signal_data_entity = self.prov_doc.entity(f"{MastProv.machine}:{k}")
                    for signal in v:
                        self.write_signal_prov(
                            signal,
                            signal_data_entity,
                            scheduler_run_activity,
                            record_shot_activity,
                            scheduler_pass_trigger_agent,
                            signals_done=signals_done,
                        )
            signals_done.append(signal)

    def write_prov(self):
        self.build_prov()

        directories = ("prov-xml", "prov-json", "prov-graph")
        for directory in directories:
            if not os.path.exists(directory):
                os.makedirs(directory)

        file_name = f"{self.data}_{self.shot}_{self.run}_prov"

        json_file = utilities.slugify(file_name) + ".json"
        xml_file = utilities.slugify(file_name) + ".xml"
        json_path = os.path.join("prov-json", json_file)
        xml_path = os.path.join("prov-xml", xml_file)
        self.prov_doc.serialize(json_path)
        self.prov_doc.serialize(xml_path, format="xml")

        if self.graph:
            dot = prov_to_dot(self.prov_doc)
            graph_file = utilities.slugify(file_name) + ".png"
            graph_path = os.path.join("prov-graph", graph_file)
            dot.write_png(graph_path)


@MastProvFactory.register_subclass("data_file")
class MastFileProv(MastProv):
    def __init__(self, params):
        super().__init__(params)
        if self.run is None:
            self.run = self.get_latest_run()
        if FREIA:
            self.file = self.file_name()
        else:
            self.file = params["data"]
        self.log_file = self.get_log()

    @staticmethod
    def validate(params) -> bool:
        client = pyuda.Client()
        if params["data"].lower()[0] in ["a", "e"]:
            return any(
                params["data"].lower() in signal.source_alias
                for signal in client.list_signals(shot=params["shot"])
            )
        else:
            return False

    def get_latest_run(self):
        signal_list = self.client.list(
            ListType.SIGNALS, shot=self.shot, alias=self.data
        )
        highest_pass_number = max([signal.pass_ for signal in signal_list])
        return highest_pass_number

    def file_name(self):
        """
        Returns the name of the datafile based on MAST convention.
        """
        file_sources = self.client.list(ListType.SOURCES, shot=self.shot)
        files = [source.filename for source in file_sources]
        file_name = next(file for file in files if self.data in file)
        if file_name:
            return file_name
        else:
            raise FileNotFoundError

    def get_scheduler_run(self, file_code=None, log_file=None):
        """
        Searches the log file for the line that contains the scheduler
        launch script and the execution start/end times.
        """
        if not file_code:
            file_code = self.data
        if not log_file:
            log_file = self.log_file
        matches = re.findall(f"run_{file_code}.*?:[0-9][0-9]:[0-9][0-9]", log_file)
        start_call = matches[0].split()
        end_call = matches[-1].split()
        scheduler_script = start_call[0]
        start_time = f"{start_call[-2]} {start_call[-1]}"
        end_time = f"{end_call[-2]} {end_call[-1]}"
        scheduler_run = (scheduler_script, start_time, end_time)
        return scheduler_run

    def get_log(self, file_code=None):
        """
        Retrieves the log file for the signal as a string.
        """
        if not file_code:
            file_code = self.data
        filename = f"{file_code}_0{self.shot}.log"
        path = f"/net/raidsrvr/data/MAST_Data/{self.shot}/Pass{self.run}/{filename}"

        if FREIA:
            log = self.client.get_text(path)
        else:
            signal = f'TESTPLUGIN::nathan(filename="{path}")'
            log = self.client.get(signal, "", raw=True).str

        return log

    def get_analysis_code(self, file_code=None):
        """
        Retrieves the analysis codes used to produce the data.
        """
        if not file_code:
            file_code = self.data
        filename = f"info_{file_code}_0{self.shot}_{self.run}.dat"
        location = f"{MastProv.data_location}/{self.shot}/Pass{self.run}/"
        return location + filename

    def build_prov(self):
        """
        Collates the provenance information for the given data file into a
        W3C-PROV document.
        """
        print(f"Generating a provenance document for {self.file} file...")

        self.prov_doc.add_namespace(MastProv.machine, MastProv.namespace_uri)

        file_entity = self.prov_doc.entity(f"{MastProv.machine}:{self.file}")

        data_ro_agent = self.prov_doc.agent(f"{MastProv.machine}:{self.data_ro()}")
        self.prov_doc.wasAttributedTo(file_entity, data_ro_agent)
        scheduler_run = self.get_scheduler_run()
        scheduler_run_activity = self.prov_doc.activity(
            f"{MastProv.machine}:{scheduler_run[0]} run={self.run}",
            startTime=scheduler_run[1],
            endTime=scheduler_run[2],
        )
        scheduler_pass_trigger = self.get_scheduler_trigger()
        scheduler_pass_trigger_agent = self.prov_doc.agent(
            f"{MastProv.machine}:{scheduler_pass_trigger}"
        )
        if FREIA:
            shot_date, shot_time = self.client.get_shot_date_time(shot=self.shot)
            start_time = f"{shot_date} {shot_time}"
        else:
            start_time = None

        record_shot_activity = self.prov_doc.activity(
            f"{MastProv.machine}: Record shot {self.shot}",
            startTime=start_time,
        )
        session_leader = self.get_session_leader()
        session_leader_agent = self.prov_doc.agent(
            f"{MastProv.machine}:{session_leader}"
        )

        input_files = self.get_input_files()
        signals_done = []
        for signal in input_files:
            self.write_signal_prov(
                signal,
                file_entity,
                scheduler_run_activity,
                record_shot_activity,
                scheduler_pass_trigger_agent,
                signals_done=signals_done,
            )
            signals_done.append(signal)

        self.prov_doc.wasAssociatedWith(record_shot_activity, session_leader_agent)
        self.prov_doc.wasGeneratedBy(file_entity, scheduler_run_activity)
        analysis_code = self.get_analysis_code()
        analysis_code_entity = self.prov_doc.entity(
            f"{MastProv.machine}:{analysis_code}"
        )

        analysis_code_author = self.get_analysis_code_author()
        analysis_code_author_agent = self.prov_doc.agent(
            f"{MastProv.machine}:{analysis_code_author}"
        )

        self.prov_doc.used(scheduler_run_activity, analysis_code_entity)
        self.prov_doc.wasAssociatedWith(
            scheduler_run_activity, scheduler_pass_trigger_agent
        )
        self.prov_doc.wasAttributedTo(analysis_code_entity, analysis_code_author_agent)


@MastProvFactory.register_subclass("uda_signal")
class MastSignalProv(MastProv):
    def __init__(self, params):
        super().__init__(params)
        if self.run == None:
            self.signal = self.client.get(self.data, self.shot)
            self.run = int(self.signal.meta["pass"])
        else:
            self.signal = self.client.get(self.data, self.shot, self.run)
        try:
            self.signal_name = self.signal.meta["signal_name"].decode("utf-8")
        except:
            self.signal_name = self.signal.meta["signal_name"]
        self.file_prov = MastFileProv(
            {
                "data": self.signal.meta["filename"][:3].decode("utf-8"),
                "shot": self.shot,
                "run": self.run,
                "graph": self.graph,
            }
        )

    @staticmethod
    def validate(params) -> bool:
        client = pyuda.Client()
        signal = params["data"]
        if signal.lower() in set(
            [signal.source_alias for signal in client.list_signals(shot=params["shot"])]
        ):
            return False
        elif signal.lower() in set(
            [
                signal.generic_name.lower()
                for signal in client.list_signals(shot=params["shot"])
            ]
        ):
            return True
        elif signal.lower() in set(
            [
                signal.signal_name.lower()
                for signal in client.list_signals(shot=params["shot"])
            ]
        ):
            return True
        else:
            return False

    def build_prov(self):
        """
        Collates the provenance information for the given signal into a
        W3C-PROV compliant json, xml and optionally a graphical output.
        """
        print(
            f"Generating a provenance document for {self.signal_name} data from shot {self.shot}, pass {self.run}..."
        )
        self.file_prov.build_prov()
        self.prov_doc = self.file_prov.prov_doc
        signal = f"{MastProv.machine}:{self.shot}_{self.run}_{self.signal_name}"
        signal_entity = self.prov_doc.entity(signal)
        self.prov_doc.wasDerivedFrom(
            signal_entity, f"{MastProv.machine}:{self.file_prov.file}"
        )


@MastProvFactory.register_subclass("image_file")
class MastImageProv:
    def __init__(self, params):
        self.prov_doc = ProvDocument()
        self.client = pyuda.Client()
        self.client.set_property("get_meta", True)
        self.data = params["data"]
        self.shot = params["shot"]
        self.run = params["run"]
        self.graph = params["graph"]
        self.ipx = self.client.get_images(
            self.data, self.shot, first_frame=0, last_frame=0
        )

    @staticmethod
    def validate(params) -> bool:
        client = pyuda.Client()
        try:
            client.get_images(params["data"], params["shot"])
        except pyuda.cpyuda.ServerException:
            return False
        else:
            return True

    def build_prov(self):
        self.prov_doc.add_namespace(MastProv.machine, MastProv.namespace_uri)

        ipx_entity = self.prov_doc.entity(f"{MastProv.machine}: shot {self.shot} camera data")

        ipx_RO = self.prov_doc.agent(f"{MastProv.machine}: IMAGE DATA RO")
        self.prov_doc.wasAttributedTo(ipx_entity, ipx_RO)

        camera_entity = self.prov_doc.entity(f"{MastProv.machine}: {self.ipx.camera}")
        record_image_data_activity = self.prov_doc.activity(f"{MastProv.machine}: Record image data", startTime=self.ipx.date_time)
        self.prov_doc.used(record_image_data_activity, camera_entity)
        self.prov_doc.wasGeneratedBy(ipx_entity, record_image_data_activity)

        if FREIA:
            shot_date, shot_time = self.client.get_shot_date_time(shot=self.shot)
            start_time = f"{shot_date} {shot_time}"
        else:
            start_time = None

        run_shot_activity = self.prov_doc.activity(
            f"{MastProv.machine}: Run shot {self.shot}",
            startTime=start_time,
        )
        session_leader = "SESSION LEADER"
        session_leader_agent = self.prov_doc.agent(
            f"{MastProv.machine}:{session_leader}"
        )

        self.prov_doc.wasAssociatedWith(run_shot_activity, session_leader_agent)
        self.prov_doc.used(run_shot_activity, camera_entity)

    def write_prov(self):
        self.build_prov()

        directories = ("prov-xml", "prov-json", "prov-graph")
        for directory in directories:
            if not os.path.exists(directory):
                os.makedirs(directory)

        file_name = f"{self.data}_{self.shot}_prov"

        json_file = utilities.slugify(file_name) + ".json"
        xml_file = utilities.slugify(file_name) + ".xml"
        json_path = os.path.join("prov-json", json_file)
        xml_path = os.path.join("prov-xml", xml_file)
        self.prov_doc.serialize(json_path)
        self.prov_doc.serialize(xml_path, format="xml")

        if self.graph:
            dot = prov_to_dot(self.prov_doc)
            graph_file = utilities.slugify(file_name) + ".png"
            graph_path = os.path.join("prov-graph", graph_file)
            dot.write_png(graph_path)


def parse_params(*args, **kwargs) -> Dict:
    """
    Collates the user-provided parameters into a dictionary of the type that the classes are expecting.

    Returns
    -------
    Dict
        A dictionary with keys: data, shot, run and graph.
    """
    params = {"data": None, "shot": None, "run": None, "graph": False}
    if args:
        if len(args) == 1 and isinstance(args[0], Mapping):
            for key, _ in params.items():
                if key in args[0]:
                    params[key] = args[0][key]
        else:
            args_dict = dict(zip_longest(params, args))
            params.update(args_dict)

    for key, _ in params.items():
        if key in kwargs:
            params[key] = kwargs[key]

    return params


def guess_data_type(params: Dict[str, Any]) -> str:
    """Runs each classes validator on the params in turn until a valid class is identified."""
    if not FREIA:
        return "uda_signal"
    for data_type in MastProvFactory.subclasses:
        if MastProvFactory.subclasses[data_type].validate(params):
            return data_type
    else:
        print(f"Invalid data type {params['data']}")
        sys.exit(1)  # TODO: Define and raise custom exception


def write_provenance(*args, **kwargs) -> None:
    params = parse_params(*args, **kwargs)
    data_type = guess_data_type(params)
    prov_data = MastProvFactory.create(data_type, params)
    prov_data.write_prov()


def main():
    parser = argparse.ArgumentParser(
        description="Retrieve and document the provenance of MAST and MAST-U " "data."
    )
    parser.add_argument("shot", help="Shot number", type=int)
    parser.add_argument(
        "run",
        nargs="?",
        default=None,
        help="Run number. Will default to the latest run.",
        type=int,
    )
    parser.add_argument("data", help="Data or signal to document provenance for.")
    parser.add_argument(
        "--graph",
        "-g",
        help="In addition to text formats, write provenance in graph form as .png. "
        "Default: false",
        action="store_true",
    )
    args = parser.parse_args()

    params = {
        "data": args.data,
        "shot": args.shot,
        "run": args.run,
        "graph": args.graph,
    }

    write_provenance(params)
